import logging
from collections import Counter
from dataclasses import dataclass, field
from datetime import datetime
from typing import Iterator

from dateparser import parse

from monopoly.constants.date import ISO8601
from monopoly.enums import RegexEnum
from monopoly.pdf import PdfPage

logger = logging.getLogger(__name__)


@dataclass
class DateMatch:
    """Helper class that holds information about a match, including the span
    (start pos and end pos), the parsed date, the raw date, the line and
    page number, as well as the entire line that was matched
    """

    span: tuple[int, int]
    parsed_date: datetime
    raw_date: str
    page_number: int
    line_number: int
    line: str


@dataclass
class DatePattern:
    regex: RegexEnum
    num_matches: int = 0
    matches: list[DateMatch] = field(default_factory=list)
    spans: list[tuple] = field(default_factory=list)

    def add_occurrence(self, matches: list[DateMatch]):
        """Add an occurrence and update the span list."""
        for match in matches:
            self.num_matches += 1
            self.matches.append(match)
            self.spans.append(match.span)

    @property
    def name(self) -> str:
        return self.regex.name.lower()

    @property
    def unique_spans(self):
        return set(self.spans)

    @property
    def span_occurrences(self) -> Counter:
        """Counts the number of occurrences per span"""
        spans = Counter(self.spans)
        return spans

    @property
    def matched_lines(self) -> list[str] | None:
        """Shows lines that are possible transactions"""
        lines_with_dates = []
        for match in self.matches:
            if match.span in self.unique_spans:
                lines_with_dates.append(match.line)

        return lines_with_dates or None


# pylint: disable=too-many-instance-attributes
class PatternMatcher:
    """Holds date regex patterns used by the generic statement handler"""

    def __iter__(self) -> Iterator[DatePattern]:
        iso8601 = [e.lower() for e in ISO8601._member_names_]
        attr_names = [name for name in dir(self) if name in iso8601]
        for attr_name in attr_names:
            attr = getattr(self, attr_name)
            if isinstance(attr, DatePattern):
                yield attr

    def __init__(self, pages: list[PdfPage]):
        self.pages = pages
        self.set_date_patterns()
        self.get_matches()

    def set_date_patterns(self):
        self.dd_mm = DatePattern(ISO8601.DD_MM)
        self.dd_mm_yy = DatePattern(ISO8601.DD_MM_YY)
        self.dd_mmm = DatePattern(ISO8601.DD_MMM)
        self.dd_mmm_yy = DatePattern(ISO8601.DD_MMM_YY)
        self.dd_mmm_yyyy = DatePattern(ISO8601.DD_MMM_YYYY)
        self.mmmm_dd_yyyy = DatePattern(ISO8601.MMMM_DD_YYYY)
        self.mmm_dd = DatePattern(ISO8601.MMM_DD)
        self.mmm_dd_yyyy = DatePattern(ISO8601.MMM_DD_YYYY)

    def get_matches(self):
        """Iterates over all possible date patterns, and retrieves transactions lines
        based on the assumption that a transaction line will always start with a date"""
        logger.debug("Iterating over patterns: %s", [p.name for p in self])
        for pattern in self:
            logger.debug("Searching for date matches for pattern %s", pattern.name)
            for i, page in enumerate(self.pages):
                for j, line in enumerate(page.lines):
                    if matches := self._extract_match(pattern, line, i, j):
                        pattern.add_occurrence(matches)

    def _extract_match(
        self, pattern: DatePattern, line: str, page_num: int, line_num: int
    ) -> list[DateMatch]:
        matches = []
        for date_match in pattern.regex.finditer(line):
            raw_date = date_match.group()
            parsed_date = parse(raw_date)
            if parsed_date:
                matches.append(
                    DateMatch(
                        span=(date_match.start(), date_match.end()),
                        parsed_date=parsed_date,
                        raw_date=raw_date,
                        page_number=page_num,
                        line_number=line_num,
                        line=date_match.string,
                    )
                )
                logger.debug(
                    "Date match found at page %s line %s: raw_date='%s' parsed_date='%s'",
                    page_num,
                    line_num,
                    raw_date,
                    parsed_date,
                )
        return matches

    def get_transaction_pattern(self) -> DatePattern:
        """
        Finds the pattern that has the highest occurrence of spans over several lines.
        For example, a line like '01 Oct  <description>' has a span of (0, 6) - we then
        check for repeats of this span, and sum it over all lines.

        Patterns ending with "yy" are analyzed first, to avoid cases where we have
        a tiebreaker with dd_mm and dd_mm_yy variants.
        """
        max_span_occurrences = 0
        most_common_pattern = None

        # Sort patterns so that those ending with "yy" come last
        sorted_patterns = sorted(self, key=lambda p: p.name.endswith("yy"))

        for pattern in sorted_patterns:
            if counter := pattern.span_occurrences:
                for _, num_occurrences in counter.most_common(2):
                    if num_occurrences >= max_span_occurrences:
                        most_common_pattern = pattern
                        max_span_occurrences = num_occurrences

        if most_common_pattern:
            logger.debug(
                "Pattern '%s' matched %s times",
                most_common_pattern.name,
                max_span_occurrences,
            )

        if not most_common_pattern:
            raise ValueError("Unable to detect most common pattern")

        return most_common_pattern

    def get_transaction_spans(self, pattern: DatePattern) -> set:
        """
        Given a pattern, returns the most probable spans based on occurrences.
        If there are two spans with the same number of occurrences e.g. (0, 6) has 17
        occurrences and (11, 17) has 17 occurrences, we return both
        spans since there is a strong likelihood that one is a transaction date,
        and the other is a posting date.
        """
        max_span_occurrences = 0
        most_common_spans = set()

        counter = pattern.span_occurrences
        for span, num_occurrences in counter.most_common(2):
            if num_occurrences >= max_span_occurrences:
                max_span_occurrences = num_occurrences
                most_common_spans.add(span)

        # Check for additional spans with occurrences >= 50% of the max span occurrences
        if len(most_common_spans) == 1:
            for span, num_occurrences in counter.items():
                if span not in most_common_spans:
                    if num_occurrences > max_span_occurrences * 0.5:
                        most_common_spans.add(span)
                        logger.debug("Adding additional span: %s", span)
                        break

        logger.debug("Most common span(s): %s", most_common_spans)

        if len(most_common_spans) > 2:
            raise ValueError("More than two date spans found in statement")

        return most_common_spans

    def get_statement_date_pattern(self) -> str:
        """
        Creates a regex pattern for the statement date based on the first statement
        date.
        """
        patterns_with_year = [p for p in self if p.name.endswith("yy")]
        yyyy_matches = {}
        for pattern in patterns_with_year:
            if pattern.matches:
                yyyy_matches[pattern.name] = pattern.matches

        if not yyyy_matches:
            raise RuntimeError(
                "No lines with `yy` or `yyyy` dates "
                "- unable to create statement date pattern"
            )

        # try yyyy patterns first before yy
        sorted_patterns: list[str] = sorted(
            yyyy_matches.keys(), key=lambda p: not p.endswith("yyyy")
        )
        unsorted_lines = []

        for pattern_name in sorted_patterns:
            lines = yyyy_matches[pattern_name]
            for line in lines:
                unsorted_lines.append((pattern_name, line))

        sorted_lines: list[tuple[str, DateMatch]] = sorted(
            unsorted_lines, key=lambda x: (x[1].page_number, x[1].line_number)
        )

        _, date_match = sorted_lines[0]
        statement_date = date_match.raw_date
        logger.debug("Found statement date: %s", statement_date)

        # assume the statement date is the first parsable YYYY date
        return statement_date



from dataclasses import dataclass, field
from typing import Optional, Pattern

from monopoly.constants import BankNames, EntryType, InternalBankNames
from monopoly.enums import RegexEnum
from monopoly.identifiers import MetadataIdentifier


@dataclass
class DateOrder:
    """
    Supported `dateparser` DATE_ORDER arguments can be found here:
    https://dateparser.readthedocs.io/en/latest/settings.html#date-order
    """

    date_order: str

    @property
    def settings(self):
        return {"DATE_ORDER": self.date_order}


# pylint: disable=too-many-instance-attributes
@dataclass(kw_only=True)
class StatementConfig:
    """
    Base configuration class storing configuration values for debit and
    credit card statements

    - `transaction_pattern` refers to the regex pattern used to capture transactions,
    where a pattern like:
        "(?P<transaction_date>\\d+/\\d+)\\s*"
        "(?P<description>.*?)\\s*"
        "(?P<amount>[\\d.,]+)$"
    is used to capture a transaction like:
        06/07 URBAN TRANSIT CO. SINGAPORE SG  1.38
    - `transaction_date_order` represents the datetime format that a specific bank uses
    for transactions. For example, "DMY" will parse 01/02/2024 as 1 Feb 2024.
    Defaults to DMY.
    - `multiline_transactions` controls whether Monopoly tries to concatenate
    transactions that are split across two lines
    - `header_pattern` is a regex pattern that is used to find the 'header' line
    of a statement, and determine if it is a debit or credit card statement.
    """

    bank_name: BankNames | InternalBankNames
    statement_type: EntryType
    transaction_pattern: Pattern[str] | RegexEnum
    statement_date_pattern: Pattern[str] | RegexEnum
    header_pattern: Pattern[str] | RegexEnum
    transaction_date_order: DateOrder = field(default_factory=lambda: DateOrder("DMY"))
    statement_date_order: DateOrder = field(default_factory=lambda: DateOrder("DMY"))
    multiline_transactions: bool = False
    has_withdraw_deposit_column: bool = False
    prev_balance_pattern: Optional[Pattern[str] | RegexEnum] = None


@dataclass
class PdfConfig:
    """Stores PDF configuration values for the `PdfParser` class

    - `password`: The password used to unlock the PDF (if it is locked)
    - `page_range`: A slice representing which pages to process. For
    example, a range of (1, -1) will mean that the first and last pages
    are skipped.
    - `page_bbox`: A tuple representing the bounding box range for every
    page. This is used to avoid weirdness like vertical text, and other
    PDF artifacts that may affect parsing.
    - `ocr_identifiers`: Applies OCR on PDFs with a specific metadata identifier.
    """

    page_range: tuple[Optional[int], Optional[int]] = (None, None)
    page_bbox: Optional[tuple[float, float, float, float]] = None
    ocr_identifiers: Optional[list[MetadataIdentifier]] = None



import logging
from re import compile as regex

from monopoly.config import StatementConfig
from monopoly.constants import (
    BankNames,
    CreditTransactionPatterns,
    DebitTransactionPatterns,
    EntryType,
    StatementBalancePatterns,
)
from monopoly.identifiers import MetadataIdentifier, TextIdentifier

from ..base import BankBase

logger = logging.getLogger(__name__)


class Ocbc(BankBase):
    credit_config = StatementConfig(
        statement_type=EntryType.CREDIT,
        bank_name=BankNames.OCBC,
        statement_date_pattern=regex(r"(\d{2}\-\d{2}\-\d{4})"),
        header_pattern=regex(r"(TRANSACTION DATE.*DESCRIPTION.*AMOUNT)"),
        prev_balance_pattern=StatementBalancePatterns.OCBC,
        transaction_pattern=CreditTransactionPatterns.OCBC,
    )

    debit_config = StatementConfig(
        statement_type=EntryType.DEBIT,
        bank_name=BankNames.OCBC,
        statement_date_pattern=regex(r"TO\s(\d+\s[A-Za-z]{3}\s\d{4})"),
        header_pattern=regex(r"(Withdrawal.*Deposit.*Balance)"),
        transaction_pattern=DebitTransactionPatterns.OCBC,
        multiline_transactions=True,
    )

    identifiers = [
        [
            MetadataIdentifier(
                creator="pdfgen", producer="Streamline PDFGen for OCBC Group"
            ),
            TextIdentifier("OCBC"),
        ],
    ]

    statement_configs = [debit_config, credit_config]



import logging
from dataclasses import dataclass
from functools import cached_property, lru_cache
from io import BytesIO
from pathlib import Path
from typing import TYPE_CHECKING, Optional, Type

import pdftotext
from pydantic import SecretStr
from pydantic_settings import BaseSettings, SettingsConfigDict
from pymupdf import TEXTFLAGS_TEXT, Document, Page

from monopoly.identifiers import MetadataIdentifier

if TYPE_CHECKING:
    from monopoly.banks import BankBase

logger = logging.getLogger(__name__)


class MissingOCRError(Exception):
    """
    Error that is raised when PDF does not contain any selectable text
    """


class PdfPasswords(BaseSettings):
    """
    Pydantic model that automatically populates variables from a .env file,
    or an environment variable called `PDF_PASSWORDS`.
    e.g. export PDF_PASSWORDS='["password123", "secretpass"]'
    """

    pdf_passwords: list[SecretStr] = [SecretStr("")]
    model_config = SettingsConfigDict(env_file=".env", extra="allow")


@dataclass
class PdfPage:
    """
    Dataclass representation of a bank statement PDF page.
    Contains the raw text of a PDF page, and allows access
    to the raw text as a list via the `lines` property
    """

    raw_text: str

    @cached_property
    def lines(self) -> list[str]:
        return self.raw_text.split("\n")


class WrongPasswordError(Exception):
    """Exception raised when an incorrect password is provided"""


class MissingPasswordError(Exception):
    """Exception raised when the document is encrypted, but no password is provided"""


class BadPasswordFormatError(Exception):
    """Exception raised passwords are not provided in a proper format"""


class PdfDocument(Document):
    """Handles logic related to the opening, unlocking, and storage of a PDF document."""

    def __init__(
        self,
        file_path: Optional[Path] = None,
        file_bytes: Optional[bytes | BytesIO] = None,
        passwords: Optional[list[SecretStr]] = None,
    ):
        self.file_path = file_path
        self.file_bytes = file_bytes
        self.passwords = passwords or PdfPasswords().pdf_passwords

        if not any([self.file_path, self.file_bytes]):
            raise RuntimeError("Either `file_path` or `file_bytes` must be passed")

        if self.file_path and self.file_bytes:
            raise RuntimeError(
                "Only one of `file_path` or `file_bytes` should be passed"
            )

        args = {"filename": self.file_path, "stream": self.file_bytes}
        super().__init__(**args)

        if self.is_encrypted:
            self._unlock_document()

        self.metadata_identifier = MetadataIdentifier(**self.metadata)

    def _unlock_document(self):
        """Attempt to unlock the document using the provided passwords."""
        if not self.is_encrypted:
            return self

        if not self.passwords:
            raise MissingPasswordError("No password found in PDF configuration")

        if not any(str(password) for password in self.passwords):
            raise MissingPasswordError("Password in PDF configuration is empty")

        if not isinstance(self.passwords, list):
            raise BadPasswordFormatError("Passwords should be stored in a list")

        if not all(isinstance(item, SecretStr) for item in self.passwords):
            raise BadPasswordFormatError("Passwords should be stored as SecretStr")

        for password in self.passwords:
            if self.authenticate(password.get_secret_value()):
                logger.debug("Successfully authenticated with password")
                return self

        raise WrongPasswordError(f"Could not open document: {self.name}")

    @cached_property
    def raw_text(self) -> str:
        """Extracts and returns the text from the PDF"""
        raw_text = ""
        for page in self:
            raw_text += page.get_text()
        return raw_text


class PdfParser:
    def __init__(
        self,
        bank: Type["BankBase"],
        document: PdfDocument,
    ):
        """
        Class responsible for parsing PDFs and returning raw text

        The page_range variable determines which pages are extracted.
        All pages are extracted by default.
        """
        self.bank = bank
        self.document = document
        self.metadata_identifier = document.metadata_identifier

    @property
    def pdf_config(self):
        return self.bank.pdf_config

    @cached_property
    def page_range(self):
        return slice(*self.pdf_config.page_range)

    @cached_property
    def page_bbox(self):
        return self.pdf_config.page_bbox

    @cached_property
    def ocr_available(self):
        if ids := self.pdf_config.ocr_identifiers:
            for identifiers in ids:
                if self.metadata_identifier.matches(identifiers):
                    return True
        return False

    @lru_cache
    def get_pages(self) -> list[PdfPage]:
        logger.debug("Extracting text from PDF")
        document = self.document

        num_pages = list(range(document.page_count))
        document.select(num_pages[self.page_range])

        if cropbox := self.page_bbox:
            logger.debug(
                "Will crop pages with crop box %s and remove vertical text", cropbox
            )
        for page in document:
            if self.page_bbox:
                page.set_cropbox(cropbox)
            page = self._remove_vertical_text(page)

        # certain statements requsire garbage collection, so that duplicate objects
        # do not cause pdftotext to fail due to missing xrefs/null values
        # however, setting `garbage=2` may cause issues with other statements
        # so an initial attempt should be made to run using `garbage=0`
        garbage_values = [0, 2]

        for garbage in garbage_values:
            try:
                pdf_byte_stream = BytesIO(document.tobytes(garbage=garbage))
                pdf = pdftotext.PDF(pdf_byte_stream, physical=True)

                # assume PDF is missing OCR if text is less than 10 chars on every page
                if all(len(page) < 10 for page in pdf):
                    raise MissingOCRError("No selectable text found")

                return [PdfPage(page) for page in pdf]
            except pdftotext.Error:
                continue
        raise RuntimeError("Unable to retrieve pages")

    @staticmethod
    def _remove_vertical_text(page: Page):
        """Helper function to remove vertical text, based on writing direction (wdir).

        This helps avoid situations where the PDF is oddly parsed, due to vertical text
        inside the PDF.

        An example of vertical text breaking a transaction:
        ```
        'HEALTHY HARVEST CAFÉ SINGAPORE SG',
        'Co Reg No: 123456',
        '10 NOV 9.80',
        ```

        Note:
            The 'dir' key represents the tuple (cosine, sine) for the angle.
            If line["dir"] != (1, 0), the text of its spans is rotated.

        """
        for block in page.get_text("dict", flags=TEXTFLAGS_TEXT)["blocks"]:
            for line in block["lines"]:
                writing_direction = line["dir"]
                if writing_direction != (1, 0):
                    page.add_redact_annot(line["bbox"])
        page.apply_redactions(images=0)
        return page

    @staticmethod
    def apply_ocr(document: PdfDocument) -> PdfDocument:
        # pylint: disable=import-outside-toplevel
        try:
            from ocrmypdf import Verbosity, configure_logging, ocr
            from ocrmypdf.exceptions import PriorOcrFoundError, TaggedPDFError
        except ImportError:
            logger.warning("ocrmypdf not installed, skipping OCR")
            return document

        added_ocr = False
        try:
            logger.debug("Applying OCR")
            original_metadata = document.metadata
            output_bytes = BytesIO()
            configure_logging(Verbosity.quiet)
            logging.getLogger("ocrmypdf").setLevel(logging.ERROR)
            ocr(
                BytesIO(document.tobytes()),
                output_bytes,
                language="eng",
                tesseract_config="tesseract.cfg",
                progress_bar=False,
                optimize=0,
                fast_web_view=999999,
                output_type="pdf",
            )
            output_bytes.seek(0)
            added_ocr = True

        except (PriorOcrFoundError, TaggedPDFError):
            pass

        # pylint: disable=attribute-defined-outside-init
        if added_ocr:
            logger.debug("Adding OCR layer to document")
            document = PdfDocument(file_bytes=output_bytes)
            document.metadata = original_metadata
        return document



import logging
from functools import cached_property
from typing import Type

from monopoly.banks import BankBase
from monopoly.config import StatementConfig
from monopoly.constants import EntryType, InternalBankNames
from monopoly.handler import StatementHandler
from monopoly.pdf import PdfPage

from .generic import DatePatternAnalyzer

logger = logging.getLogger(__name__)


class GenericBank(BankBase):
    """
    Empty bank class with variables that can be populated by
    the `GenericStatementHandler` class
    """

    # populated by generic statement handler
    @property
    def statement_configs(self):
        return None

    @property
    def identifiers(self):
        return None


class GenericStatementHandler(StatementHandler):
    def __init__(self, bank: Type[BankBase], pages: list[PdfPage]):
        self.analyzer = DatePatternAnalyzer(pages)
        bank.statement_configs = list(
            filter(None, [self.debit_config, self.credit_config])
        )
        super().__init__(bank, pages)

    # override get_header and ignore passed config, since
    # the header line has already been found
    def get_header(self, _: StatementConfig):
        return self.header_pattern

    @cached_property
    def debit_config(self):
        if self.statement_type == EntryType.DEBIT:
            logger.debug("Creating debit statement config")

            return StatementConfig(
                statement_type=EntryType.DEBIT,
                bank_name=InternalBankNames.GENERIC,
                transaction_pattern=self.transaction_pattern,
                statement_date_pattern=self.statement_date_pattern,
                multiline_transactions=self.multiline_transactions,
                header_pattern=self.header_pattern,
            )
        return None

    @cached_property
    def credit_config(self):
        if self.statement_type == EntryType.CREDIT:
            logger.debug("Creating credit statement config")

            return StatementConfig(
                statement_type=EntryType.CREDIT,
                bank_name=InternalBankNames.GENERIC,
                prev_balance_pattern=self.prev_balance_pattern,
                transaction_pattern=self.transaction_pattern,
                statement_date_pattern=self.statement_date_pattern,
                header_pattern=self.header_pattern,
                multiline_transactions=self.multiline_transactions,
            )
        return None

    @cached_property
    def transaction_pattern(self):
        return self.analyzer.create_transaction_pattern()

    @cached_property
    def statement_type(self):
        return self.analyzer.get_statement_type()

    @cached_property
    def statement_date_pattern(self):
        return self.analyzer.create_statement_date_pattern()

    @cached_property
    def multiline_transactions(self):
        return self.analyzer.check_if_multiline()

    @cached_property
    def header_pattern(self):
        lines = self.analyzer.lines_before_first_transaction
        return self.analyzer.get_debit_statement_header_line(lines)

    @cached_property
    def prev_balance_pattern(self):
        return self.analyzer.create_previous_balance_regex()



import re
from enum import Enum
from functools import cached_property
from typing import Any

from strenum import StrEnum


# pylint: disable=unused-argument,no-self-argument
class AutoEnum(StrEnum):
    """Generates lower case values for enums
    e.g. CITIBANK -> citibank
    """

    def _generate_next_value_(name: str, *_):  # type: ignore
        return name.lower()


class RegexEnum(Enum):
    """
    A subclass of `Enum` that converts member to a compiled
    regular expression pattern.
    """

    def __new__(cls, pattern):
        """
        Create a new `RegexEnum` member.

        Args:
            pattern (str): A regular expression pattern.
        """
        if not isinstance(pattern, str):
            raise TypeError(f"Pattern must be a string, not {type(pattern)}")

        obj = object.__new__(cls)
        obj._value_ = pattern
        return obj

    @cached_property
    def regex(self):
        return re.compile(self.value)

    def __str__(self):
        return self.value

    def match(self, value) -> re.Match:
        """
        Wrapper for re.match
        """
        return self.regex.search(value)

    def findall(self, value) -> list[Any]:
        """
        Wrapper for re.findall
        """
        return self.regex.findall(value)

    def finditer(self, value) -> list[re.Match]:
        """
        Wrapper for re.finditer
        """
        return self.regex.finditer(value)

    def search(self, value) -> re.Match:
        """
        Wrapper for re.search
        """
        return self.regex.search(value)



import logging
from re import compile as regex

from monopoly.config import PdfConfig, StatementConfig
from monopoly.constants import (
    BankNames,
    CreditTransactionPatterns,
    EntryType,
    StatementBalancePatterns,
)
from monopoly.identifiers import MetadataIdentifier, TextIdentifier

from ..base import BankBase

logger = logging.getLogger(__name__)


class StandardChartered(BankBase):
    credit_config = StatementConfig(
        statement_type=EntryType.CREDIT,
        bank_name=BankNames.STANDARD_CHARTERED,
        statement_date_pattern=regex(r"(\d{2}\s\w+\s\d{4})"),
        header_pattern=regex(r"(Transaction.*Posting.*Amount)"),
        prev_balance_pattern=StatementBalancePatterns.STANDARD_CHARTERED,
        transaction_pattern=CreditTransactionPatterns.STANDARD_CHARTERED,
    )

    pdf_config = PdfConfig(
        page_range=(0, -1),
        page_bbox=(0, 0, 580, 820),
    )

    identifiers = [
        [
            MetadataIdentifier(
                title="eStatement",
                producer="iText",
            ),
            TextIdentifier("Standard Chartered"),
        ]
    ]

    statement_configs = [credit_config]



import logging
import re

from monopoly.constants import EntryType
from monopoly.statements.transaction import Transaction, TransactionGroupDict

from .base import BaseStatement, SafetyCheckError

logger = logging.getLogger(__name__)


class CreditStatement(BaseStatement):
    """
    A dataclass representation of a credit statement
    """

    statement_type = EntryType.CREDIT

    def post_process_transactions(self, transactions) -> list[Transaction]:
        previous_month_balances = self.get_prev_month_balances()
        if previous_month_balances:
            first_transaction_date = next(iter(transactions)).date
            for prev_month_balance in previous_month_balances:
                groupdict = TransactionGroupDict(**prev_month_balance.groupdict())
                groupdict.transaction_date = first_transaction_date
                prev_month_transaction = Transaction(**groupdict)
                transactions.insert(0, prev_month_transaction)
        return transactions

    def get_prev_month_balances(self) -> list[re.Match]:
        """
        Returns the previous month's statement balance as a transaction,
        if it exists in the statement.

        The date is later replaced with a more accurate date by the statement handler.
        """
        prev_balances = []

        if pattern := self.config.prev_balance_pattern:
            for page in self.pages:
                for line in page.lines:
                    if match := pattern.search(line):
                        prev_balances.append(match)

        return prev_balances

    def perform_safety_check(self) -> bool:
        """Checks that the total sum of all transactions is present
        somewhere within the document

        Text is re-extracted from the page, as some bank-specific bounding-box
        configurations (e.g. HSBC) may preclude the total from being extracted

        Returns `False` if the total does not exist in the document.
        """
        numbers = self.get_all_numbers_from_document()
        transactions = self.transactions

        amounts = [transaction.amount for transaction in transactions]
        total_amount = abs(round(sum(amounts), 2))

        result = total_amount in numbers
        if not result:
            raise SafetyCheckError(
                f"Total amount {total_amount} cannot be found in credit statement"
            )

        return result



import logging
from re import compile as regex

from monopoly.config import StatementConfig
from monopoly.constants import (
    BankNames,
    CreditTransactionPatterns,
    DebitTransactionPatterns,
    EntryType,
    StatementBalancePatterns,
)
from monopoly.identifiers import MetadataIdentifier

from ..base import BankBase

logger = logging.getLogger(__name__)


class Maybank(BankBase):
    debit_config = StatementConfig(
        statement_type=EntryType.DEBIT,
        bank_name=BankNames.MAYBANK,
        statement_date_pattern=regex(r"(?:結單日期)[:\s]+(\d{2}\/\d{2}\/\d{2})"),
        header_pattern=regex(r"(DATE.*DESCRIPTION.*AMOUNT.*BALANCE)"),
        transaction_pattern=DebitTransactionPatterns.MAYBANK,
        has_withdraw_deposit_column=False,
        multiline_transactions=True,
    )

    credit_config = StatementConfig(
        statement_type=EntryType.CREDIT,
        bank_name=BankNames.MAYBANK,
        statement_date_pattern=regex(r"(\d{2}\s[A-Z]{3}\s\d{2})"),
        header_pattern=regex(r"(Date.*Description.*Amount)"),
        transaction_pattern=CreditTransactionPatterns.MAYBANK,
        prev_balance_pattern=StatementBalancePatterns.MAYBANK,
        multiline_transactions=True,
    )

    identifiers = [
        [
            MetadataIdentifier(
                author="Maybank2U.com",
                creator="Maybank2u.com",
                producer="iText",
            ),
        ],
        [
            MetadataIdentifier(
                title="Credit Card Statement",
                author="Maybank2U.com",
                producer="iText",
            ),
        ],
    ]

    statement_configs = [debit_config, credit_config]



import logging
from functools import lru_cache
from typing import Type

from monopoly.banks import BankBase
from monopoly.config import StatementConfig
from monopoly.constants import EntryType
from monopoly.pdf import PdfPage
from monopoly.statements import BaseStatement, CreditStatement, DebitStatement

logger = logging.getLogger(__name__)


class StatementHandler:
    """
    Retrieves statement information like transactions from the PDF,
    and identifies the statement as either a debit or credit statement
    based on the debit and credit config.
    """

    def __init__(self, bank: Type[BankBase], pages: list[PdfPage]):
        self.bank = bank
        self.pages = pages

    def get_header(self, config: StatementConfig) -> str | None:
        pattern = config.header_pattern

        for page in self.pages:
            for line in page.lines:
                if match := pattern.search(line):
                    return match.group().lower()
        return None

    @lru_cache
    def get_statement(self) -> BaseStatement:
        pages = self.pages
        for config in self.bank.statement_configs:
            if header := self.get_header(config):
                match config.statement_type:
                    case EntryType.DEBIT:
                        logger.debug("Statement type detected: %s", EntryType.DEBIT)
                        return DebitStatement(pages, config, header)
                    case EntryType.CREDIT:
                        logger.debug("Statement type detected: %s", EntryType.CREDIT)
                        return CreditStatement(pages, config, header)

        raise RuntimeError("Could not create statement object")



import logging
from dataclasses import Field, fields
from functools import cached_property
from typing import TYPE_CHECKING, Any, Type

from monopoly.identifiers import Identifier, TextIdentifier
from monopoly.pdf import PdfDocument

if TYPE_CHECKING:
    from .base import BankBase

logger = logging.getLogger(__name__)


class BankDetector:
    def __init__(self, document: PdfDocument):
        self.document = document

    @cached_property
    def metadata_items(self) -> list[Any]:
        """
        Retrieves encryption and metadata identifiers from a bank statement PDF
        """
        identifiers: list[Identifier] = []
        if metadata_identifier := self.document.metadata_identifier:
            identifiers.append(metadata_identifier)

        if not identifiers:
            raise ValueError("Could not get identifier")

        return identifiers

    def detect_bank(self, banks: list[Type["BankBase"]]) -> Type["BankBase"] | None:
        """
        Reads the encryption metadata or actual metadata (if the PDF is not encrypted),
        and checks for a bank based on unique identifiers.
        """
        if not banks:
            banks = []

        logger.debug("Found PDF properties: %s", self.metadata_items)

        for bank in banks:
            if self.is_bank_identified(bank):
                return bank
        return None

    def is_bank_identified(
        self,
        bank: Type["BankBase"],
    ) -> bool:
        """
        Checks if a bank is identified based on a list of metadata items.
        """
        for grouped_identifiers in bank.identifiers:  # type: ignore
            text_identifiers = list(
                filter(lambda i: isinstance(i, TextIdentifier), grouped_identifiers)
            )
            pdf_property_identifiers = list(
                filter(lambda i: not isinstance(i, TextIdentifier), grouped_identifiers)
            )

            if pdf_property_identifiers:
                if len(self.metadata_items) != len(pdf_property_identifiers):
                    continue

                if not self.pdf_properties_match(pdf_property_identifiers):
                    continue

                if not self.text_identifiers_match(text_identifiers):
                    logger.warning("PDF metadata matches but text not found")
                    return False

                logger.debug("Identified statement bank: %s", bank.__name__)
                return True

            if text_identifiers:
                if self.text_identifiers_match(text_identifiers):
                    return True

        return False

    def text_identifiers_match(self, text_identifiers: list[TextIdentifier]) -> bool:
        if not text_identifiers:
            return True

        for identifier in text_identifiers:
            if identifier.text not in self.document.raw_text:
                return False

        logger.debug("Text identifier found in PDF")
        return True

    def pdf_properties_match(
        self,
        grouped_identifiers: list[Identifier],
    ) -> bool:
        """
        Checks if all identifiers in the group match at least one metadata item.
        """
        for identifier in grouped_identifiers:
            matching_metadata = [
                metadata
                for metadata in self.metadata_items
                if type(metadata) is type(identifier)
            ]

            if not any(
                self.check_matching_identifier(metadata, identifier)
                for metadata in matching_metadata
            ):
                return False
        return True

    def check_matching_identifier(
        self,
        metadata: Identifier,
        identifier: Identifier,
    ) -> bool:
        """
        Checks if all fields in the metadata match the corresponding identifier fields.
        """
        return all(
            self.check_matching_field(field, metadata, identifier)
            for field in fields(metadata)
        )

    def check_matching_field(
        self,
        dataclass_field: Field,
        metadata: Identifier,
        identifier: Identifier,
    ) -> bool:
        """
        Checks if a field in the metadata matches the corresponding identifier field.
        """
        field_value = getattr(metadata, dataclass_field.name)
        identifier_value = getattr(identifier, dataclass_field.name)

        # if identifier is empty, we assume a match
        # this means only the identifiers in the bank
        # class need to be matched.
        if not identifier_value:
            return True

        # support partial string matching
        if isinstance(field_value, str) and isinstance(identifier_value, str):
            return identifier_value in field_value and identifier_value != ""

        # if not a string, only support exact match
        return identifier_value == field_value



import logging
from re import compile as regex

from monopoly.config import StatementConfig
from monopoly.constants import (
    BankNames,
    CreditTransactionPatterns,
    DebitTransactionPatterns,
    EntryType,
    StatementBalancePatterns,
)
from monopoly.identifiers import MetadataIdentifier, TextIdentifier

from ..base import BankBase

logger = logging.getLogger(__name__)


class Dbs(BankBase):
    credit_config = StatementConfig(
        statement_type=EntryType.CREDIT,
        bank_name=BankNames.DBS,
        statement_date_pattern=regex(r"(\d{2}\s[A-Za-z]{3}\s\d{4})"),
        multiline_transactions=False,
        header_pattern=regex(r"(DATE.*DESCRIPTION.*AMOUNT)"),
        transaction_pattern=CreditTransactionPatterns.DBS,
        prev_balance_pattern=StatementBalancePatterns.DBS,
    )

    debit_config = StatementConfig(
        statement_type=EntryType.DEBIT,
        bank_name=BankNames.DBS,
        statement_date_pattern=regex(r"(\d{2}\s[A-Za-z]{3}\s\d{4})"),
        multiline_transactions=True,
        header_pattern=regex(r"(WITHDRAWAL.*DEPOSIT.*BALANCE)"),
        transaction_pattern=DebitTransactionPatterns.DBS,
    )

    identifiers = [
        [
            TextIdentifier("DBS"),
            MetadataIdentifier(creator="Quadient CXM AG"),
        ],
    ]

    statement_configs = [debit_config, credit_config]
